"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.register = void 0;
const transforms_1 = require("@volar/transforms");
const node_1 = require("vscode-languageserver/node");
const path = require("upath");
const shared_1 = require("@volar/shared");
const shared_2 = require("@vue/shared");
const scriptAst_1 = require("../parsers/scriptAst");
function register({ sourceFiles, tsLs, ts, vueHost }) {
    return (item, newOffset) => {
        const data = item.data;
        if (!data)
            return item;
        const sourceFile = sourceFiles.get(data.uri);
        if (!sourceFile)
            return item;
        if (data.mode === 'ts') {
            return getTsResult(sourceFile, item, data);
        }
        if (data.mode === 'html') {
            return getHtmlResult(sourceFile, item, data);
        }
        if (data.mode === 'autoImport') {
            return getAutoImportResult(sourceFile, item, data);
        }
        return item;
        function getTsResult(sourceFile, vueItem, data) {
            for (const sourceMap of sourceFile.getTsSourceMaps()) {
                if (sourceMap.mappedDocument.uri !== data.docUri)
                    continue;
                let newTsOffset;
                if (newOffset) {
                    for (const tsRange of sourceMap.getMappedRanges2(newOffset)) {
                        if (!tsRange.data.capabilities.completion)
                            continue;
                        newTsOffset = tsRange.start;
                        break;
                    }
                }
                data.tsItem = tsLs.doCompletionResolve(data.tsItem, newTsOffset);
                const newVueItem = transforms_1.transformCompletionItem(data.tsItem, tsRange => sourceMap.getSourceRange(tsRange.start, tsRange.end));
                newVueItem.data = data;
                // TODO: this is a patch for import ts file icon
                if (newVueItem.detail !== data.tsItem.detail + '.ts') {
                    newVueItem.detail = data.tsItem.detail;
                }
                return newVueItem;
            }
            return vueItem;
        }
        function getHtmlResult(sourceFile, vueItem, data) {
            var _a, _b;
            let tsItem = data.tsItem;
            if (!tsItem)
                return vueItem;
            tsItem = tsLs.doCompletionResolve(tsItem);
            vueItem.tags = [...(_a = vueItem.tags) !== null && _a !== void 0 ? _a : [], ...(_b = tsItem.tags) !== null && _b !== void 0 ? _b : []];
            const details = [];
            const documentations = [];
            if (vueItem.detail)
                details.push(vueItem.detail);
            if (tsItem.detail)
                details.push(tsItem.detail);
            if (details.length) {
                vueItem.detail = details.join('\n\n');
            }
            if (vueItem.documentation)
                documentations.push(typeof vueItem.documentation === 'string' ? vueItem.documentation : vueItem.documentation.value);
            if (tsItem.documentation)
                documentations.push(typeof tsItem.documentation === 'string' ? tsItem.documentation : tsItem.documentation.value);
            if (documentations.length) {
                vueItem.documentation = {
                    kind: node_1.MarkupKind.Markdown,
                    value: documentations.join('\n\n'),
                };
            }
            return vueItem;
        }
        function getAutoImportResult(sourceFile, vueItem, data) {
            const importFile = shared_1.uriToFsPath(data.importUri);
            const rPath = path.relative(vueHost.getCurrentDirectory(), importFile);
            let importPath = path.relative(path.dirname(data.uri), data.importUri);
            if (!importPath.startsWith('.')) {
                importPath = './' + importPath;
            }
            vueItem.labelDetails = { qualifier: rPath };
            const descriptor = sourceFile.getDescriptor();
            const scriptImport = descriptor.script ? getLastImportNode(descriptor.script.content, descriptor.script.lang) : undefined;
            const scriptSetupImport = descriptor.scriptSetup ? getLastImportNode(descriptor.scriptSetup.content, descriptor.scriptSetup.lang) : undefined;
            const componentName = shared_2.capitalize(shared_2.camelize(vueItem.label));
            const textDoc = sourceFile.getTextDocument();
            let insertText = '';
            const planAResult = planAInsertText();
            if (planAResult) {
                insertText = planAResult.insertText;
                vueItem.detail = planAResult.description + '\n\n' + rPath;
            }
            else {
                insertText = planBInsertText();
                vueItem.detail = `Auto import from '${importPath}'\n\n${rPath}`;
            }
            if (descriptor.scriptSetup) {
                vueItem.additionalTextEdits = [
                    node_1.TextEdit.insert(textDoc.positionAt(descriptor.scriptSetup.loc.start + (scriptSetupImport ? scriptSetupImport.end : 0)), '\n' + insertText),
                ];
            }
            else if (descriptor.script) {
                vueItem.additionalTextEdits = [
                    node_1.TextEdit.insert(textDoc.positionAt(descriptor.script.loc.start + (scriptImport ? scriptImport.end : 0)), '\n' + insertText),
                ];
                const scriptAst = scriptAst_1.parse(ts, descriptor.script.content, descriptor.script.lang, true, true);
                const exportDefault = scriptAst.exportDefault;
                if (exportDefault) {
                    // https://github.com/microsoft/TypeScript/issues/36174
                    const printer = ts.createPrinter();
                    if (exportDefault.componentsOption && exportDefault.componentsOptionNode) {
                        exportDefault.componentsOptionNode.properties.push(ts.factory.createShorthandPropertyAssignment(componentName));
                        const printText = printer.printNode(ts.EmitHint.Expression, exportDefault.componentsOptionNode, scriptAst.sourceFile);
                        vueItem.additionalTextEdits.push(node_1.TextEdit.replace(node_1.Range.create(textDoc.positionAt(descriptor.script.loc.start + exportDefault.componentsOption.start), textDoc.positionAt(descriptor.script.loc.start + exportDefault.componentsOption.end)), unescape(printText.replace(/\\u/g, '%u'))));
                    }
                    else if (exportDefault.args && exportDefault.argsNode) {
                        exportDefault.argsNode.properties.push(ts.factory.createShorthandPropertyAssignment(`components: { ${componentName} }`));
                        const printText = printer.printNode(ts.EmitHint.Expression, exportDefault.argsNode, scriptAst.sourceFile);
                        vueItem.additionalTextEdits.push(node_1.TextEdit.replace(node_1.Range.create(textDoc.positionAt(descriptor.script.loc.start + exportDefault.args.start), textDoc.positionAt(descriptor.script.loc.start + exportDefault.args.end)), unescape(printText.replace(/\\u/g, '%u'))));
                    }
                }
            }
            return vueItem;
            function planAInsertText() {
                const scriptUrl = sourceFile.getVirtualScriptUri();
                if (!scriptUrl)
                    return;
                const tsImportName = shared_2.camelize(path.basename(importFile).replace(/\./g, '-'));
                const tsDetail = tsLs.__internal__.raw.getCompletionEntryDetails(shared_1.uriToFsPath(scriptUrl), 0, tsImportName, {}, importFile, undefined, undefined);
                if (tsDetail === null || tsDetail === void 0 ? void 0 : tsDetail.codeActions) {
                    for (const action of tsDetail.codeActions) {
                        for (const change of action.changes) {
                            for (const textChange of change.textChanges) {
                                if (textChange.newText.indexOf(`import ${tsImportName} `) >= 0) {
                                    return {
                                        insertText: textChange.newText.replace(`import ${tsImportName} `, `import ${componentName} `).trim(),
                                        description: action.description,
                                    };
                                }
                            }
                        }
                    }
                }
            }
            function planBInsertText() {
                const anyImport = scriptSetupImport !== null && scriptSetupImport !== void 0 ? scriptSetupImport : scriptImport;
                let withSemicolon = true;
                let quote = '"';
                if (anyImport) {
                    withSemicolon = anyImport.text.endsWith(';');
                    quote = anyImport.text.includes("'") ? "'" : '"';
                }
                return `import ${componentName} from ${quote}${importPath}${quote}${withSemicolon ? ';' : ''}`;
            }
        }
    };
    function getLastImportNode(code, lang) {
        const sourceFile = ts.createSourceFile('foo.' + lang, code, ts.ScriptTarget.Latest);
        let importNode;
        sourceFile.forEachChild(node => {
            if (ts.isImportDeclaration(node)) {
                importNode = node;
            }
        });
        return importNode ? {
            text: importNode.getFullText(sourceFile).trim(),
            end: importNode.getEnd(),
        } : undefined;
    }
}
exports.register = register;
//# sourceMappingURL=completionResolve.js.map